/////////////////////////////////////////////////////////////////////
//
//  File:    REPORT.CPP
//
//  Author:  Scott Wood
//
//  Created: 01 Nov 92 RSW
//
//  Revised: 27 Jan 94 RSW
//           29 Mar 94 RSW, ported to Windows from XVT
//           01 Jul 94 RSW, converted to Borland OWL
//
//  Purpose: To provide the Windows print interface necessary for sending
//           reports to the printer.
//           Reports are generated by the following structure:
//             Declare Report instance:      Report rpt;
//             Open Report:                  rpt.Open (doSetup, "Test Report");
//               where doSetup is true or
//                 false depending on user.
//             Begin Report:                 rpt.BeginReport ("Test 1");
//               This allows multiple reports
//               within a single Device context.
//             Print report using either
//              defined routines or the DC.  rpt.printf ("Hello World\n");
//                                           rpt.TextOut (TPoint, "Hi");
//                                           rpt.PDC->TextOut (TPoint, "Hi");
//             End Report:                   rpt.EndReport ();
//             Close Report:                 rpt.Close ();
//
//  Notes:   This report class is designed to allow generation of reports
//            without having to worry about page formatting, banding, etc.
//            Currently it supports only a printer context or a file.
//            Print preview has not been added.
//
/////////////////////////////////////////////////////////////////////
#include <iostream.h>
#include <report.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <strfuncs.h>
#include <time.h>
#include <rptlib.h>
#include <rptlib.rh>
#include <cntrwin.h>

DlgReportAbort *Report::pAbortWin;

/////////////////////////////////////////////////////////////////////
// Function:    Report
// Purpose:     Report Constructor
// Arguments:   none
/////////////////////////////////////////////////////////////////////
Report::Report (void)
	 : pInReport (false)
	 , pFtrFunc (0)
	 , pHdrFunc (0)
  {
    pData = new DlgReportSetup::ReportData;
	 DlgReportSetup (0, *pData).GetDefaultPrinter ();
  };

Report::~Report (void)
  {
    delete pData;
  }

/////////////////////////////////////////////////////////////////////
// Function: pReportAbortProc
// Purpose:  Report Abort Procedure, called internally
/////////////////////////////////////////////////////////////////////
BOOL FAR PASCAL Report::pReportAbortProc (HDC, int)
  {
	 TApplication *app = GetApplicationObject ();
	 app->PumpWaitingMessages ();
	 return !pAbortWin->IsAbort ();
  };

/////////////////////////////////////////////////////////////////////
// Function:    Open
// Purpose:     Create a Report Context
// Arguments:   boolean doSetup - true if user setup desired
//              string title    - title of report
/////////////////////////////////////////////////////////////////////
boolean Report::Open (boolean doSetup, const string _FAR & title)
  {
	 pTitle = title;

	 if (doSetup)
		if (!SetupReport ())
		  return false;

	 if ((pAbortWin = new DlgReportAbort (0, title.c_str (),
													  pData->GetDeviceName (),
													  pData->GetOutputName ())) == 0)
		{
		  delete pHDC;
		  return false;
		};
	 pAbortWin->Create ();
	 pAbortWin->ShowActivity ("Initializing...");

	 switch (pData->ReportOutput)
		{
		  case reportOutputPrinter:
			 pHDC = new TPrintDC (pData->GetDriverName (), pData->GetDeviceName (),
										 pData->GetOutputName (), pData->GetDevMode ());
			 if (!pHDC)
				return false;

			 pHDC->SetAbortProc (pReportAbortProc);

			 pHDC->SelectStockObject (ANSI_FIXED_FONT);

			 pPageWd = pHDC->GetDeviceCaps (HORZRES);
			 pPageHt = pHDC->GetDeviceCaps (VERTRES);
			 pVRes   = pHDC->GetDeviceCaps (LOGPIXELSY);
			 pHRes   = pHDC->GetDeviceCaps (LOGPIXELSX);

			 pSetMapping ();

			 pPageWd = pPageWd * (1440.0 / (float)pHRes);
			 pPageHt = pPageHt * (1440.0 / (float)pVRes);
			 pVRes   = 1440;
			 pHRes   = 1440;

			 if (pHDC->StartDoc (pTitle.c_str (), 0) < 0)
				{
				  Close ();
				  return false;
				};
			 break;

		  case reportOutputFile:
			 pOutputChno = _lcreat (pData->OutputFile.Path ().c_str (), 0);

			 pPageWd = 256;
			 pPageHt = 65;
			 pVRes   = 10;
			 pHRes   = 5;

			 pLineHt = 1;
			 pCharWd = 1;

			 break;

		};

	 pFirstPage = true;
	 pInReport = true;
	 return true;
  };

/////////////////////////////////////////////////////////////////////
// Function:     Close
// Purpose:      Close a Report Context
// Arguments:    None
/////////////////////////////////////////////////////////////////////
void Report::Close (void)
  {
    if (!pAbortWin)
      return;
    
	 pAbortWin->ShowActivity ("Closing Report");

	 switch (pData->ReportOutput)
		{
		  case reportOutputPrinter:
			 pHDC->EndDoc ();

			 pAbortWin->Destroy ();
			 delete pHDC;
			 break;

		  case reportOutputFile:
			 _lclose (pOutputChno);
			 break;
		};

	 delete pAbortWin;
	 pInReport = false;
    pAbortWin = 0;
  };

/////////////////////////////////////////////////////////////////////
// Function:     SetupReport
// Purpose:      User Setup of Report Parameters
// Arguments:    None
/////////////////////////////////////////////////////////////////////
boolean Report::SetupReport (void)
  {
	 pData->Flags = 0;
	 return (DlgReportSetup (0, *pData).Execute () == IDOK);
  };

/////////////////////////////////////////////////////////////////////
// Function:     SetupPrinter
// Purpose:      User setup of particular Printer
// Arguments:    None
/////////////////////////////////////////////////////////////////////
boolean Report::SetupPrinter (void)
  {
	 pData->Flags = PD_PRINTSETUP;
	 return (DlgReportSetup (0, *pData).Execute () == IDOK);
  };

/////////////////////////////////////////////////////////////////////
// Function:     RedirectOutput
// Purpose:      Allows forcing of report to particular destination
// Arguments:    where    - Printer or File
//               fileName - if File, destination name
/////////////////////////////////////////////////////////////////////
boolean Report::RedirectOutput (reportOutputEnum where, const FileName _FAR & fileName)
  {
	 pData->ReportOutput = where;
	 pData->OutputFile	= fileName;
	 return true;
  };

/////////////////////////////////////////////////////////////////////
// Function:     OrientPaper
// Purpose:      Force particular page orientation
// Arguments:    orientVal - paperOrientPortrait or paperOrientLandscape
/////////////////////////////////////////////////////////////////////
void Report::OrientPaper (paperOrientEnum orientVal)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 DEVMODE dm = *pData->GetDevMode ();
	 dm.dmOrientation = (int)orientVal;
	 pData->SetDevMode (&dm);
  };

/////////////////////////////////////////////////////////////////////
// Function:     BeginReport
// Purpose:      For each report within a Report Context, use a
//                BeginReport / EndReport pair.
// Arguments:    reportName - name of this report
/////////////////////////////////////////////////////////////////////
boolean Report::BeginReport (const string _FAR & reportName)
  {
	 SetFont ("Courier", 10, 0, FW_NORMAL, false);

	 pBorderRct.top    = pVRes * pData->Margins.top / 100;
	 pBorderRct.left   = pHRes * pData->Margins.left / 100;
	 pBorderRct.bottom = pPageHt - (pVRes * pData->Margins.bottom / 100);
	 pBorderRct.right  = pPageWd - (pHRes * pData->Margins.right / 100);

	 pMarginRct = pBorderRct;

	 SetX (0);
	 SetY (32000);   // force first print to pagebreak

	 pTestMargin = true;

	 pTestPage  = true;
	 pFirstPage = true;
	 pLastPage  = false;

	 pPageno = 0;

	 char buf[80];
	 sprintf (buf, "Printing %s", reportName.c_str ());
	 pAbortWin->ShowActivity (buf);

	 pInReport = true;
	 return true;
  };

////////////////////////////////////////////////////////////////////////
// Function:   EndReport
// Purpose:    End an individual report
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::EndReport (void)
  {
	 pLastPage = true;
	 PageBreak ();
  }

////////////////////////////////////////////////////////////////////////
// Function:   ChooseFont
// Purpose:    Allow user to select report font
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::ChooseFont (void)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 CHOOSEFONT cf;
	 LOGFONT logFont;

	 memset (&cf, 0, sizeof(CHOOSEFONT));
	 cf.lStructSize = sizeof(CHOOSEFONT);
	 cf.hwndOwner   = 0;
	 cf.lpLogFont   = &logFont;
	 cf.Flags       = CF_PRINTERFONTS;
	 cf.hDC         = (HDC)pHDC;
	 cf.nFontType   = PRINTER_FONTTYPE;

	 if (::ChooseFont (&cf))
		SetFont (logFont.lfFaceName, logFont.lfHeight, logFont.lfWidth,
					logFont.lfWeight, logFont.lfItalic);
  };

////////////////////////////////////////////////////////////////////////
// Function:   GetFont
// Purpose:    Get the current font characteristics
// Arguments:  typeFace  - Font Name
//             height    - Font Height
//             width     - Font Width
//             weight    - Intensity
//             italic    - Is italic
////////////////////////////////////////////////////////////////////////
void Report::GetFont (string _FAR & typeFace, int& height, int& width, int& weight, boolean& italic)
  {
	 typeFace = pLogFont.lfFaceName;
	 height   = pLogFont.lfHeight;
	 width    = pLogFont.lfWidth;
	 weight   = pLogFont.lfWeight;
	 italic   = pLogFont.lfItalic;
  };

////////////////////////////////////////////////////////////////////////
// Function:   SetFont
// Purpose:    Sets the font type and size for printing.
// Arguments:  typeFace - Font Name
//             height   - height
//             width    - Font Width, if 0, will be determined
//             weight   - FW_NORMAL, FW_BOLD
//             italic   - true or false
////////////////////////////////////////////////////////////////////////
void Report::SetFont (const string _FAR & typeFace, int height, int width, int weight, boolean italic)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 pLogFont.lfHeight = height * 20;
	 pLogFont.lfWidth  = width;
	 pLogFont.lfEscapement  = 0;
	 pLogFont.lfOrientation = 0;
	 pLogFont.lfWeight = weight;
	 pLogFont.lfItalic = italic;
	 pLogFont.lfUnderline = 0;
	 pLogFont.lfStrikeOut = 0;
	 pLogFont.lfCharSet = ANSI_CHARSET;
	 pLogFont.lfOutPrecision = OUT_DEFAULT_PRECIS;
	 pLogFont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	 pLogFont.lfQuality = PROOF_QUALITY;
	 pLogFont.lfPitchAndFamily = VARIABLE_PITCH | FF_ROMAN;

	 strcpy (pLogFont.lfFaceName, typeFace.c_str ());

	 pSetFont ();
  };

////////////////////////////////////////////////////////////////////////
// Function:   SetFontHeight
// Purpose:    Set a new font height
// Arguments:  height - new height
////////////////////////////////////////////////////////////////////////
void Report::SetFontHeight (int height)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 pLogFont.lfHeight = height * 20;
	 pLogFont.lfWidth  = 0;
	 pSetFont ();
  };

////////////////////////////////////////////////////////////////////////
// Function:   SetFontWeight
// Purpose:    Set a new font weight
// Arguments:  weight - FW_NORMAL, FW_BOLD, see Windows documentation
////////////////////////////////////////////////////////////////////////
void Report::SetFontWeight (int weight)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 pLogFont.lfWeight = weight;
	 pSetFont ();
  };

////////////////////////////////////////////////////////////////////////
// Function:   SetFontItalic
// Purpose:    Set italic mode
// Arguments:  italic - true or false
////////////////////////////////////////////////////////////////////////
void Report::SetFontItalic (boolean italic)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 pLogFont.lfItalic = italic;
	 pSetFont ();
  };

////////////////////////////////////////////////////////////////////////
// Function:   SetCharPerLine
// Purpose:    This routines allows fonts matching of a page.  If a
//              particular number of characters in desired per line, use
//              this routines to achieve the desired result.
// Arguments:  charPerLine - number of characters desired
////////////////////////////////////////////////////////////////////////
void Report::SetCharPerLine (int charPerLine)
  {
	 pCharPerLine = charPerLine;

	 if (pData->ReportOutput == reportOutputFile)
		return;

	 int lineWd = (int)(pMarginRct.right - pMarginRct.left + 1);

	 int aveWd = ((lineWd / charPerLine) * 10 + 5) / 10;

	 int ht = 0;

	 do {
		ht += 20;
		pLogFont.lfHeight = ht;
		pLogFont.lfWidth  = 0;
		pSetFont ();
	 } while (pCharWd < aveWd);

	 while (pCharWd > aveWd)
		{
		  pLogFont.lfHeight = --ht;
		  pLogFont.lfWidth  = 0;
		  pSetFont ();
		};

	 pCharPerLine = lineWd / pCharWd;
  };

////////////////////////////////////////////////////////////////////////
// Function:   printf
// Purpose:    Works like the standard printf
// Arguments:  fmt   - printf format, see C documentation
//             ...   - variable argument list
////////////////////////////////////////////////////////////////////////
int Report::printf (const char *fmt, ...)
  {
	 char buf[256];

	 va_list ap;
	 va_start (ap, fmt);
	 int cnt = vsprintf (buf, fmt, ap);
	 va_end (ap);

	 TSize size;

	 for (int i=0; buf[i]; i++)
		{
		  switch (pData->ReportOutput)
			 {
				case reportOutputPrinter:
				  switch (buf[i])
					 {
						case '\t':
						  size = pHDC->GetTextExtent (" ", 1);
						  pX += size.cx;
						  break;

						case '\r':
						  pX = (int)pMarginRct.left;
						  break;

						case '\n':
						  pX = (int)pMarginRct.left;
						  pY += pLineHt;
						  break;

						default:
						  TSize size = pHDC->GetTextExtent (&buf[i], 1);

						  if (pTestMargin)
							 {
								if (pX + size.cx > pMarginRct.right)
								  {
									 pX = (int)pMarginRct.left;
									 pY += size.cy;
								  };

								if (pY + size.cy > pMarginRct.bottom)
								  PageBreak ();
							 };

						  pHDC->TextOut (pX, -pY, &buf[i], 1);
						  pX += size.cx;
						  break;
					 };
				  break;

				case reportOutputFile:
				  switch (buf[i])
					 {
						case '\t':
						  _lwrite (pOutputChno, &buf[i], 1);
						  break;

						case '\r':
						  break;

						case '\n':
						  _lwrite (pOutputChno, "\r\n", 2);
						  pX = (int)pMarginRct.left;
						  pPutMargin ();
						  pY++;
						  break;

						default:
						  if (pTestMargin)
							 {
								if (pX > pMarginRct.right)
								  {
									 _lwrite (pOutputChno, "\r\n", 2);
									 pX = (int)pMarginRct.left;
									 pPutMargin ();
									 pY++;
								  };

								if (pY > pMarginRct.bottom)
								  PageBreak ();
							 };

						  _lwrite (pOutputChno, &buf[i], 1);
						  pX++;
						  break;
					 };
				  break;

			 };
		};

	 return cnt;
  };

////////////////////////////////////////////////////////////////////////
// Function:   printDashLine
// Purpose:    prints a line the width of the page.
// Arguments:  str   - Buffer to print
//             solid - true of solid line desired otherwise dash
////////////////////////////////////////////////////////////////////////
void Report::PrintDashLine (const char *str, boolean solid)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 if (!solid || pData->ReportOutput == reportOutputFile)
		{
		  char buf[256];

		  int width;

		  if (pData->ReportOutput == reportOutputFile)
			 width = pCharPerLine;
		  else
			 {
				TSize size;
				size = pHDC->GetTextExtent ("-", 1);
				width = (int)((pMarginRct.right - pMarginRct.left) / size.cx);
			 };

		  memset (buf, '-', width);
		  buf[width] = '\0';

		  int i, j;

		  int k = strlen (str);
		  j = (width / 2) - (k + 2) / 2;

		  if (k)
			 buf[j++] = ' ';
		  for (i = 0; str[i]; i++, j++)
			 buf[j] = str[i];
		  if (k)
			 buf[j] = ' ';

		  printf ("%s", buf);
		}
	 else
		{
		  char buf[256];

		  if (!isstrnil (str))
			 {
				sprintf (buf, " %s ", str);

				TSize size;
				size = pHDC->GetTextExtent (buf, strlen (buf));

				int begx = (int)(pMarginRct.left + ((pMarginRct.right - pMarginRct.left) / 2 - size.cx / 2));
				int endx = (int)(pMarginRct.left + ((pMarginRct.right - pMarginRct.left) / 2 + size.cx / 2));

				int y = pY + size.cy / 3;

				MoveTo (TPoint (pMarginRct.left, y));
				LineTo (TPoint (begx, y));

				TextOut (TPoint (begx, pY), buf, strlen (buf));

				MoveTo (TPoint (endx, y));
				LineTo (TPoint (pMarginRct.right, y));
			 }
		  else
			 {
				TSize size = pHDC->GetTextExtent (" ", 1);

				int y = pY + size.cy / 3;
				MoveTo (TPoint (pMarginRct.left, y));
				LineTo (TPoint (pMarginRct.right, y));
			 };
		};

	 printf ("\n");
  };

////////////////////////////////////////////////////////////////////////
// Function:  PrintLeftJustify
// Purpose:   Prints the data left justified on the line.  No newline
//             chars should be used.  After call, should call
//             printf ("\n") to add proper height line feed.
// Arguments:  str - Buffer to print
////////////////////////////////////////////////////////////////////////
int Report::PrintLeftJustify (const char *str)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 pX = (int)pMarginRct.left;
	 printf ("%s", str);

	 return strlen (str);
  };

////////////////////////////////////////////////////////////////////////
// Function:   PrintRightJustify
// Purpose:    Prints the data right justified on the line.  No newline
//             chars should be used.  After call, should call
//             printf ("\n") to add proper height line feed.
// Arguments:  str - Buffer to print
////////////////////////////////////////////////////////////////////////
int Report::PrintRightJustify (const char *str)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 if (pData->ReportOutput == reportOutputFile)
		{
		  int x = pX;

		  pX = pCharPerLine - strlen (str);
		  printf ("%*c%s", pX-x, ' ', str);
		}
	 else
		{
		  TSize size;
		  size = pHDC->GetTextExtent (str, strlen (str));

		  pX = (int)(pMarginRct.right - (size.cx + pCharWd));
		  printf ("%s", str);
		};

	 return strlen (str);
  };

////////////////////////////////////////////////////////////////////////
// Function:   PrintCenterJustify
// Purpose:    Prints the data centered on the line.  No newline
//             chars should be used.  After call, should call
//             printf ("\n") to add proper height line feed.
// Arguments:  str - Buffer to print
////////////////////////////////////////////////////////////////////////
int Report::PrintCenterJustify (const char *str)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 if (pData->ReportOutput == reportOutputFile)
		{
		  int x = pX;

		  pX = (int)(pMarginRct.left + pCharPerLine / 2 - strlen (str) / 2);
		  if (pX < pMarginRct.left)
			 pX = (int)pMarginRct.left;

		  printf ("%*c%s", pX-x, ' ', str);
		}
	 else
		{
		  TSize size;
		  size = pHDC->GetTextExtent (str, strlen (str));

		  pX = (int)(pMarginRct.left + ((pMarginRct.right - pMarginRct.left) / 2 - size.cx / 2));
		  if (pX < pMarginRct.left)
			 pX = (int)pMarginRct.left;

		  printf ("%s", str);
		};

	 return strlen (str);
  };

////////////////////////////////////////////////////////////////////////
// Function:   PrintNewLine
// Purpose:    For a linefeed.  Shortcut routine.
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::PrintNewLine (void)
  {
	 printf ("\n");
  };

////////////////////////////////////////////////////////////////////////
// Function:   PrintLine
// Purpose:    Much like dashed line routine, except print a line of
//              specified width.
// Arguments:  length - number of chars wide
////////////////////////////////////////////////////////////////////////
void Report::PrintLine (int length)
  {
	 if (TestPageBreak (0))
      PageBreak ();

    if (pData->ReportOutput == reportOutputFile)
      {
        char buf[256];

        for (int i=0; i < length && i < sizeof(buf); i++)
          buf[i] = '-';
        buf[i] = 0;

        printf ("%s", buf);
        return;
		};

	 TSize size = pHDC->GetTextExtent (" ", 1);

	 MoveTo (TPoint (pX, pY + size.cy / 3));

	 pX += pCharWd * length;
	 if (pX > pMarginRct.right)
		pX = (int)pMarginRct.right;

	 LineTo (TPoint (pX, pY + size.cy / 3));
  }

////////////////////////////////////////////////////////////////////////
// Function:   PrintChar
// Purpose:    Like PrintLine only print length number of chars (ch).
// Arguments:  ch     - Char to print
//             length - Number to print
////////////////////////////////////////////////////////////////////////
int Report::PrintChar (char ch, int length)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 char buf[256];

	 for (int i=0; i < length && i < sizeof(buf); i++)
		buf[i] = ch;
	 buf[i] = 0;

	 printf ("%s", buf);
	 return i;
  };

//////////////////////////////////////////////////////////////////////////
// Function:  TestPageBreak
// Purpose:   Determines if a page break required.  Returns 0 if not at
//             break and 1 if time to break
//
// Notes:     Will only test for a page break when the testPage flag is
//             set.  It is always set unless the footer is being processed.
//             Checking for page break in the footer will always yield
//             true and make it unable to continue processing
// Arguments: cnt - number of additional lines to add into test
//////////////////////////////////////////////////////////////////////////
boolean Report::TestPageBreak (int cnt)
  {
	 if (pTestPage && pTestMargin)
		return ((pY + pLineHt + (cnt * pLineHt)) >= pMarginRct.bottom);

	 return false;
  }

////////////////////////////////////////////////////////////////////////
// Function:  TestPercentPageLeft
// Purpose:   Test for a percentage of the remaining page.
//            Input must be an integer from 1 thru 100.
// Arguments: percent - percentage desired
////////////////////////////////////////////////////////////////////////
boolean Report::TestPercentPageLeft (int percent)
  {
    if ((percent <= 0) || (percent > 100))
      return true;

    float divisor = 100 / percent;

    return ((pY+pLineHt <= (pMarginRct.bottom - pMarginRct.bottom/divisor)));
  };

////////////////////////////////////////////////////////////////////////
// Function:  TestLinesLeft
// Purpose:   Test for a the number of lines remaining on the
//             page.  If there remains the number of lines
//             requested, true is returned.
// Arguments: lines - number of lines to check
////////////////////////////////////////////////////////////////////////
boolean Report::TestLinesLeft (int lines)
  {
	 if (NumLinesAvailable() >= lines)
		return true;
	 return false;
 };

////////////////////////////////////////////////////////////////////////
// Function:  NumLinesAvailable
// Purpose:   Returns the number of lines remaining on the page.
//
// Version:        1.00 Release 2.0
//                               //
// Date Modified:  18 Aug 93 KCS
////////////////////////////////////////////////////////////////////////
int Report::NumLinesAvailable (void)
  {
    return ((int)((pMarginRct.bottom - pY) / pLineHt));
  };

////////////////////////////////////////////////////////////////////////
// Function:  X
// Purpose:   Returns the current X position on the page.
//            This value is relative to the current margin
// Arguments: None
////////////////////////////////////////////////////////////////////////
int Report::X (void)
  {
    return (int)(pX - pMarginRct.left);
  }

////////////////////////////////////////////////////////////////////////
// Function:  Y
// Purpose:   Returns the current Y position on the page.
//            This value is relative to the current margin
// Arguments: None
////////////////////////////////////////////////////////////////////////
int Report::Y (void)
  {
    return (int)(pY - pMarginRct.top);
  }

////////////////////////////////////////////////////////////////////////
// Function:  SetX
// Purpose:   Sets a new X position on the page.
//            This value is relative to the current margin.
// Arguments: x - new value
////////////////////////////////////////////////////////////////////////
void Report::SetX (int x)
  {
    pX = (!pTestMargin ? x : (int)(x + pMarginRct.left));
  }

////////////////////////////////////////////////////////////////////////
// Function:  Y
// Purpose:   Sets a new Y position on the page.
//            This value is relative to the current margin
// Arguments: y - new value
////////////////////////////////////////////////////////////////////////
void Report::SetY (int y)
  {
    pY = (int)(y + pMarginRct.top);
    if (!pTestMargin)
      pY = y;
  }

////////////////////////////////////////////////////////////////////////
// Function:  SetMarginRct
// Purpose:   Sets the margin boundary for the page.
// Arguments: rct - defined border
////////////////////////////////////////////////////////////////////////
void Report::SetMarginRct (TRect _FAR & rct)
  {
    pMarginRct = rct;
    if (pX < rct.left)
      pX = (int)rct.left;
    if (pY < rct.top)
      pY = (int)rct.top;
  };

////////////////////////////////////////////////////////////////////////
// Function:  SetBorderRct
// Purpose:   Sets the border boundary for the page.
// Arguments: rct - defined border
////////////////////////////////////////////////////////////////////////
void Report::SetBorderRct (TRect _FAR & rct)
  {
	 pBorderRct = rct;
  };

//////////////////////////////////////////////////////////////////////////
// Function:  PageBreak
// Purpose:   If a valid page no (>0), adds the ftr if set and then
//            calls private pPageBreak to do the actual processing.  Sets
//            up the next page.
// Arguments: None
//////////////////////////////////////////////////////////////////////////
void Report::PageBreak (void)
  {
    if (!pFirstPage)
      {
        if (pFtrFunc)
          {
				pTestMargin = false;
            pTestPage   = false;  //disables page break check in ftr processing

            (*pFtrFunc) (this, pFtrData);
          };

        pTestPage   = true;  //resets the test for page break
        pTestMargin = true;

        pEndPage ();
      };

    if (!pLastPage)
      {
        pStartPage ();

        if (pHdrFunc)
          (*pHdrFunc)(this, pHdrData);
      };
  };

////////////////////////////////////////////////////////////////////////
// Function:  TextOut
// Purpose:   Passthru function for outputting text
// Arguments: pt  - location on page
//            str - what to print
//            len - length to print.  -1 = all
////////////////////////////////////////////////////////////////////////
void Report::TextOut (const TPoint _FAR & pt, const char *str, int len)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 if (TestPageBreak (0))
		PageBreak ();

	 pHDC->TextOut ((int)pt.x, (int)-pt.y, str, len);
  };

////////////////////////////////////////////////////////////////////////
// Function:  MoveTo
// Purpose:   Passthru function for moving pen
// Arguments: pt  - location on page
////////////////////////////////////////////////////////////////////////
void Report::MoveTo (const TPoint _FAR & pt)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 if (pData->ReportOutput == reportOutputFile)
		return;

	 pHDC->MoveTo ((int)pt.x, (int)(-pt.y));
  };

////////////////////////////////////////////////////////////////////////
// Function:  LineTo
// Purpose:   Passthru function for drawing line
// Arguments: pt  - location on page
////////////////////////////////////////////////////////////////////////
void Report::LineTo (const TPoint _FAR & pt)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 if (pData->ReportOutput == reportOutputFile)
		return;

	 pHDC->LineTo ((int)pt.x, (int)(-pt.y));
  };

////////////////////////////////////////////////////////////////////////
// Function:  Rectangle
// Purpose:   Passthru function for drawing rectangle
// Arguments: rct - defined rectangle
////////////////////////////////////////////////////////////////////////
void Report::Rectangle (const TRect _FAR & rct)
  {
	 if (TestPageBreak (0))
		PageBreak ();

	 if (pData->ReportOutput == reportOutputFile)
		return;

    TRect drct (rct);
    drct.top = -drct.top;
    drct.bottom = -drct.bottom;

    pHDC->Rectangle (drct);
  }

////////////////////////////////////////////////////////////////////////
// Function:  SetPageHeader
// Purpose:   On each new page break, the function defined will be
//             called.  This routine sets up the defined function.
// Arguments: hdrFunc - Function to call
//            data    - Data to pass
////////////////////////////////////////////////////////////////////////
void Report::SetPageHeader (void (*hdrFunc)(Report *, void *), void _FAR *data)
  {
	 pHdrFunc = hdrFunc;
	 pHdrData = data;
  };

////////////////////////////////////////////////////////////////////////
// Function:  SetPageFooter
// Purpose:   Before each new page break, the function defined will be
//             called.  This routine sets up the defined function.
// Arguments: ftrFunc - Function to call
//            data    - Data to pass
////////////////////////////////////////////////////////////////////////
void Report::SetPageFooter (void (*ftrFunc)(Report *, void *), void _FAR *data)
  {
	 pFtrFunc = ftrFunc;
	 pFtrData = data;
  };

////////////////////////////////////////////////////////////////////////
// Function:  DrawBitmap
// Purpose:   Attempt to draw a Device independent bitmap onto the page.
//            Note:  This will not work on all printers.
// Arguments: hInstance - Where bitmap is found
//            name      - Name of bitmap
//            rct       - Where to place bitmap
////////////////////////////////////////////////////////////////////////
void Report::DrawBitmap (HINSTANCE hInstance, const char *name, TRect _FAR & rct)
  {
	 if (pData->ReportOutput == reportOutputFile)
		return;

	 if (TestPageBreak (0))
		PageBreak ();

	 TBitmap bitMap (hInstance, name);

	 TMemoryDC hdcMem (*pHDC);
	 hdcMem.SelectObject (bitMap);
	 hdcMem.SetMapMode (pHDC->GetMapMode ());

	 BITMAP bm;
	 bitMap.GetObject (bm);

	 TPoint ptOrg (0, 0);
	 hdcMem.DPtoLP (&ptOrg, 1);

	 TPoint ptSize (bm.bmWidth, bm.bmHeight);
	 hdcMem.DPtoLP (&ptSize, 1);

	 TPoint dstOrg ((int)rct.left, (int)-rct.top);

	 TPoint dstSize ((int)(rct.right - rct.left), (int)-(rct.bottom - rct.top));

	 pHDC->StretchBlt (dstOrg.x, dstOrg.y, dstSize.x, dstSize.y, hdcMem,
				 ptOrg.x, ptOrg.y, ptSize.x, ptSize.y, SRCCOPY);
  };

////////////////////////////////////////////////////////////////////////
// Function:   pStartPage
// Purpose:    Start a page sequence, called internally
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::pStartPage (void)
  {
    pPageno++;
	 pFirstPage = false;

    pX = (int)pMarginRct.left;
    pY = (int)pMarginRct.top;

	 switch (pData->ReportOutput)
		{
		  case reportOutputPrinter:
			 pHDC->StartPage ();
			 pSetMapping ();
			 pSetFont ();
			 break;

		  case reportOutputFile:
			 for (int i=0; i < pY; i++)
				_lwrite (pOutputChno, "\r\n", 2);
			 pPutMargin ();
			 break;
		};

	 pAbortWin->ShowPageNumber (pPageno);
  };

////////////////////////////////////////////////////////////////////////
// Function:   pSetMapping
// Purpose:    Set map mode of page
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::pSetMapping (void)
  {
	 pHDC->SetMapMode (MM_TWIPS);
  };

////////////////////////////////////////////////////////////////////////
// Function:   pEndPage
// Purpose:    End a page
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::pEndPage (void)
  {
	 switch (pData->ReportOutput)
		{
		  case reportOutputPrinter:
			 pHDC->EndPage ();
			 break;

		  case reportOutputFile:
			 _lwrite (pOutputChno, "\f", 1);
			 break;
		};
  };

////////////////////////////////////////////////////////////////////////
// Function:   pSetFont
// Purpose:    Internal setfont
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::pSetFont (void)
  {
	 pHDC->SelectObject (TFont (&pLogFont));

	 TEXTMETRIC textMetric;
	 pHDC->GetTextMetrics (textMetric);

	 pLineHt = textMetric.tmHeight;
	 pCharWd = textMetric.tmAveCharWidth;
  };

////////////////////////////////////////////////////////////////////////
// Function:   pPutMargin
// Purpose:    For file printing, prints the left margin
// Arguments:  None
////////////////////////////////////////////////////////////////////////
void Report::pPutMargin ()
  {
	 for (int i=0; i < pX; i++)
		_lwrite (pOutputChno, " ", 1);
  };

